/////////////////////////////////////////////////////////////////////////////////////////
// Module Name: Currents.cpp
// Project: Power Monitor #42
// Measurement of AC currents and phase, also Central Task Dispatcher
// by Jan Breemer
// jan@breem.nl, janbee@hack42.nl
//
// Functions in this module:
//
// ISR(TIMER3_COMPA_vect)
// UpdateCurrents ()
// InitCurrents ()
/////////////////////////////////////////////////////////////////////////////////////////

#include "Globals.h"

unsigned int CH08_Sin, CH09_Sin, CH10_Sin, CH11_Sin, CH12_Sin, CH13_Sin, CH14_Sin;
unsigned int CH08_Cos, CH09_Cos, CH10_Cos, CH11_Cos, CH12_Cos, CH13_Cos, CH14_Cos;
int CH08_Amplitude, CH09_Amplitude, CH10_Amplitude, CH11_Amplitude, CH12_Amplitude, CH13_Amplitude, CH14_Amplitude;
int CH08_Phase, CH09_Phase, CH10_Phase, CH11_Phase, CH12_Phase, CH13_Phase, CH14_Phase;
int PhSh11, PhSh10, PhSh09, PhSh08; // These phase shifts are derived from the jumper block settings.

// Calibrations of Current measurements:
// A/D converter range is 1023, and we average over 25 periods. Max will be 1023 * 25 = 25575
// A/D converter Voltage range is 0..5 Volt, which is 5 / (2 sqrt(2)) = 1.77 Volt RMS
// So we calculate the calibration factor as:
// Current = measured / 25565 * 1.77 / Rt * F
// where Rt is the termination resistor for the current transformer
// and F the conversion ratio of the current transformer.
// The * 10.0 is because we want the answer in deci-Amperes.
const float CalCh08 = 1.0 / 25565.0 * 1.77 / 56.0 * 1000.0 * 10.0;
const float CalCh09 = 1.0 / 25565.0 * 1.77 / 56.0 * 1000.0 * 10.0;
const float CalCh10 = 1.0 / 25565.0 * 1.77 / 56.0 * 1000.0 * 10.0;
const float CalCh11 = 1.0 / 25565.0 * 1.77 / 56.0 * 1000.0 * 10.0;
const float CalCh12 = 1.0 / 25565.0 * 1.77 / 56.0 * 1000.0 * 10.0; // T-phase
const float CalCh13 = 1.0 / 25565.0 * 1.77 / 56.0 * 1000.0 * 10.0; // S-phase
const float CalCh14 = 1.0 / 25565.0 * 1.77 / 56.0 * 1000.0 * 10.0; // R-phase

unsigned int SampleNumber;
byte NumberOfAverages;
bool FoundZeroCrossing, HaveMainsPower;
bool DoReadaSensor, RequestButton1Pressed, RequestButton2Pressed;
bool PrevRequestButton1Pressed, PrevRequestButton2Pressed;
bool Blink;
bool DoCalculate, DoPublishPower;

//********************************************************************************************************
ISR(TIMER3_COMPA_vect) // interrupt at 20 kHz = 50 usec interval
{
  DIAGPIN_1_H // Show the In-Service time of this interrupt routine

  static unsigned int BlinkTimer;
  if (++BlinkTimer > 3000) // 3 Hz Blink period
  {
    BlinkTimer = 0;
    Blink = !Blink;
    EvaluateLeds();
  }

  unsigned int ADresult = ADC; // ADresult is from the conversion started in the previous timeslice!!!
  static unsigned int MainsTimeOutCounter;
  static byte ReadaSensorCount, RequestButton1Counter, RequestButton2Counter;

  // This is the primary Task Dispatcher
  // The timeslices 1 .. 7, 101 .. 107, 201 .. 207 and 301.. 307 are used to
  // sample the analog signals from the current transformers.
  // Timeslices in between and after are used for other tasks
  switch (++SampleNumber)
  {
    // Positive Sine terms
    case 1:
      DIAGPIN_3_H           // DIAGPIN_3 shows the time for sampling the currents
      CH14_Sin += ADresult;
      ADMUX = 0x45;
    break;
    case 2:
      CH13_Sin += ADresult;
      ADMUX = 0x44;
    break;
    case 3:
      CH12_Sin += ADresult;
      ADMUX = 0x43;
    break;
    case 4:
      CH11_Sin += ADresult;
      ADMUX = 0x42;
    break;
    case 5:
      CH10_Sin += ADresult;
      ADMUX = 0x41;
    break;
    case 6:
      CH09_Sin += ADresult;
      ADMUX = 0x40;
    break;
    case 7:
      CH08_Sin += ADresult;
      ADMUX = 0x46;
      DIAGPIN_3_L
    break;

    case 25:
      // we sample the signal a few times to prevent reaction on a spike
      #ifdef USE_REQUEST_BOX1
      RequestButton1Counter = 0;
      if (REQUEST_1_PRESSED) RequestButton1Counter++;
      #endif
      #ifdef USE_REQUEST_BOX2
      RequestButton2Counter = 0;
      if (REQUEST_2_PRESSED) RequestButton2Counter++;
      #endif
    break;

    // Positive Cosine terms:
    case 101:
      DIAGPIN_3_H
      CH14_Cos += ADresult;
      ADMUX = 0x45;
    break;
    case 102:
      CH13_Cos += ADresult;
      ADMUX = 0x44;
    break;
    case 103:
      CH12_Cos += ADresult;
      ADMUX = 0x43;
    break;
    case 104:
      CH11_Cos += ADresult;
      ADMUX = 0x42;
    break;
    case 105:
      CH10_Cos += ADresult;
      ADMUX = 0x41;
    break;
    case 106:
      CH09_Cos += ADresult;
      ADMUX = 0x40;
    break;
    case 107:
      CH08_Cos += ADresult;
      ADMUX = 0x46;
      DIAGPIN_3_L
    break;

    case 120:
       #ifdef USE_REQUEST_BOX1
      if (REQUEST_1_PRESSED) RequestButton1Counter++;
      #endif
      #ifdef USE_REQUEST_BOX2
      if (REQUEST_2_PRESSED) RequestButton2Counter++;
      #endif
    break;

    // Negative Sine terms
    case 201:
      DIAGPIN_3_H
      CH14_Sin -= ADresult;
      ADMUX = 0x45;
    break;
    case 202:
      CH13_Sin -= ADresult;
      ADMUX = 0x44;
    break;
    case 203:
      CH12_Sin -= ADresult;
      ADMUX = 0x43;
    break;
    case 204:
      CH11_Sin -= ADresult;
      ADMUX = 0x42;
    break;
    case 205:
      CH10_Sin -= ADresult;
      ADMUX = 0x41;
    break;
    case 206:
      CH09_Sin -= ADresult;
      ADMUX = 0x40;
    break;
    case 207:
      CH08_Sin -= ADresult;
      ADMUX = 0x46;
      DIAGPIN_3_L
    break;

    case 220:
      #ifdef USE_REQUEST_BOX1
      if (REQUEST_1_PRESSED) RequestButton1Counter++;
      #endif
      #ifdef USE_REQUEST_BOX2
      if (REQUEST_2_PRESSED) RequestButton2Counter++;
      #endif
    break;

    case 230:
      if (++ReadaSensorCount > 9)
      {
        ReadaSensorCount = 0;
        DoReadaSensor = true; // in 5 Hz tempo
      }
    break;

    // Negative Cosine terms
    case 301:
      DIAGPIN_3_H
      CH14_Cos -= ADresult;
      ADMUX = 0x45;
    break;
    case 302:
      CH13_Cos -= ADresult;
      ADMUX = 0x44;
    break;
    case 303:
      CH12_Cos -= ADresult;
      ADMUX = 0x43;
    break;
    case 304:
      CH11_Cos -= ADresult;
      ADMUX = 0x42;
    break;
    case 305:
      CH10_Cos -= ADresult;
      ADMUX = 0x41;
    break;
    case 306:
      CH09_Cos -= ADresult;
      ADMUX = 0x40;
    break;
    case 307:
      CH08_Cos -= ADresult;
      ADMUX = 0x47;  // continue with channel 15 to find the next zero crossing.
      DIAGPIN_3_L
      DIAGPIN_4_H // start of 4th quadrant 
    break;
 
    case 308:
      #ifdef USE_REQUEST_BOX1
      if (REQUEST_1_PRESSED) RequestButton1Counter++;
      if (RequestButton1Counter >= 3) RequestButton1Pressed = true;
      if (RequestButton1Pressed != PrevRequestButton1Pressed)
      {
        PrevRequestButton1Pressed = RequestButton1Pressed;
        DoPublishStatus = true;
      }
      #endif
      #ifdef USE_REQUEST_BOX2
      if (REQUEST_2_PRESSED) RequestButton2Counter++;
      if (RequestButton2Counter >= 3) RequestButton2Pressed = true;
      if (RequestButton2Pressed != PrevRequestButton2Pressed)
      {
        PrevRequestButton2Pressed = RequestButton2Pressed;
        DoPublishStatus = true;
      }
      #endif
    break;

    case 309:
      if (++NumberOfAverages > 24) // we have averaged over 0.5 seconds
      {
        DoCalculate = true;
        NumberOfAverages = 0;
      }
    break;

    case 310:
      //if (!HaveMainsPower)
      // SampleNumber = 315; // skip further finding a zero crossing
    break;

    case 312:
      // We start here and wait for a signal < 450 on Channel 15
      // Channel 15 is the Voltage input from the R-phase and
      // is used only to find a positive going zero crossing.
      // We start however with finding a negative signal (below half scale)
      MainsTimeOutCounter = 0;
      ADMUX = 0x47; // next A/D result will be from channel 15
    break;

    case 313:
      #ifdef USE_CURRENTS
      // We have implemented a mechanism which lets this statemachine continue to run
      // even when there is a short mains power interruption.
      // For longer black-outs the whole unit will fail and automatically
      // restarts after the power is restored.
      if (!HaveMainsPower) break;
      if (ADresult > 450) 
      {
        SampleNumber--; // stay here until ADresult <= 450  = neagive half period
        // note the auto increment in the switch header
        if (++MainsTimeOutCounter > 30000) // 1.5 seconds
        {
          MainsTimeOutCounter = 0;
          HaveMainsPower = false;
          Serial.println("No  Mains Zero Crossing found");
          Serial.println("Falling back to default Task Dispatcher.");
          Serial.println();
        }
      }
      #endif
    break;

    case 314:  // now wait for ADresult > 512
      #ifdef USE_CURRENTS
      if (HaveMainsPower) 
      {
        if (ADresult < 512)
        {
          SampleNumber--; // stay here until a positive zero crossing is found
          if (++MainsTimeOutCounter > 30000) // 1.5 seconds
          {
            MainsTimeOutCounter = 0;
            HaveMainsPower = false;
            Serial.println("No Mains Zero Crossing found");
            Serial.println("Falling back to default Task Dispatcher.");
            Serial.println();
          }
        }
        else // we found a positive going zero crossing
        {
          ADMUX = 0x46;     // next ADresult will be from channel 14
          SampleNumber = 0;
          DIAGPIN_4_L       // end of 4th quadrant
          DIAGPIN_2_T       // 50 Hz trigger on pin A7
        }
      }  
      #endif  
    break;

    #ifdef USE_CURRENTS
    // This code snippet is for the situation where we have USE_CURRENTS defined
    // but the mains power is temporary absent.
    // This code detects that mains power is back again
    case 317:
      ADMUX = 0x47; // next A/D result will be from channel 15
    break;

    case 318:
      if ((ADresult < 400) || (ADresult > 600)) // Mains Power is back ?
      {
        SampleNumber = 311;
        HaveMainsPower = true;
      }
    break;
    #endif
    
    case 399:
      // if we do not HaveMainsPower or USE_CURRENTS is not defined we'll end up here
      // and repeat the cycle in approximately 50 Hz tempo
      DIAGPIN_4_L    // end of (so to say) 4th quadrant
      DIAGPIN_2_T    // 50 Hz trigger on pin A7
      SampleNumber = 0;
    break;
    
  } // switch()
  
  ADCSRA |= 0x40; // Start a new A/D conversion
  DIAGPIN_1_L
}

//********************************************************************************************************
void UpdateCurrents (void) // called very often by FW_Monitor/loop()
{
#ifdef USE_CURRENTS
if (HaveMainsPower)
  {
  if (DoCalculate) // set at 0.5 seconds interval, when 25 mains periods have passed
  // These calculations take around 3.5 msec, so we cannot do it inside the interrupt routine.
  // The signal DoCalculate is given just after the 4th quadrant is sampled, so there is enough time left.
    {
      DoCalculate = false;
      // We calculate the absolute amplitude as the square root of the sum of the squared Cosine and Sine terms.
      // We calculate the phase angle as the atan2 of (Sineterm, Cosineterm).
      // atan2() gives the correct answer for the whole circle 0 .. 2pi.
      // We convert the angles from radians to decidegrees to pass integer values in the messages.
      // The angles are corrected for the 120 and 240 degrees phase shift of the S-and T-phases
      // and the channel timing error of n * 50us / 20ms * 360 deg = 0.9 deg per timeslice.
      // There is also some phase shift in the transformer which provides the voltage on channel 15.
      // We add 3600 decidegrees and apply the modulo 3600 operation to prevent negative angles.
      // In the end the values are published as degrees.

      const int TransformerPhaseShift = -24;  // decidegrees
      PhSh11 = PhSh10 = PhSh09 = PhSh08 = 0;
      if (Ch08Mode == S) PhSh08 = 2400; if (Ch08Mode == T) PhSh08 = 1200;
      if (Ch09Mode == S) PhSh09 = 2400; if (Ch09Mode == T) PhSh09 = 1200;
      if (Ch10Mode == S) PhSh10 = 2400; if (Ch10Mode == T) PhSh10 = 1200;
      if (Ch11Mode == S) PhSh11 = 2400; if (Ch11Mode == T) PhSh11 = 1200;

      CH14_Amplitude = round(sqrt(float(CH14_Sin) * float(CH14_Sin) + float(CH14_Cos) * float(CH14_Cos)) * CalCh14);
      CH14_Phase = round(atan2(CH14_Sin, CH14_Cos) * 1800 / pi + 3600 + TransformerPhaseShift - 9) % 3600;

      CH13_Amplitude = round(sqrt(float(CH13_Sin) * float(CH13_Sin) + float(CH13_Cos) * float(CH13_Cos)) * CalCh13);
      CH13_Phase = round(atan2(CH13_Sin, CH13_Cos) * 1800 / pi + 3600  + TransformerPhaseShift - 2400 - 18) % 3600;

      CH12_Amplitude = round(sqrt(float(CH12_Sin) * float(CH12_Sin) + float(CH12_Cos) * float(CH12_Cos)) * CalCh12);
      CH12_Phase = round(atan2(CH12_Sin, CH12_Cos) * 1800 / pi + 3600  + TransformerPhaseShift - 1200 - 27) % 3600;

      CH11_Amplitude = round(sqrt(float(CH11_Sin) * float(CH11_Sin) + float(CH11_Cos) * float(CH11_Cos)) * CalCh11);
      CH11_Phase = round(atan2(CH11_Sin, CH11_Cos) * 1800 / pi + 3600  + TransformerPhaseShift - PhSh11 - 20) % 3600;

      CH10_Amplitude = round(sqrt(float(CH10_Sin) * float(CH10_Sin) + float(CH10_Cos) * float(CH10_Cos)) * CalCh10);
      CH10_Phase = round(atan2(CH10_Sin, CH10_Cos) * 1800 / pi + 3600  + TransformerPhaseShift - PhSh10 - 15) % 3600;

      CH09_Amplitude = round(sqrt(float(CH09_Sin) * float(CH09_Sin) + float(CH09_Cos) * float(CH09_Cos)) * CalCh09);
      CH09_Phase = round(atan2(CH09_Sin, CH09_Cos) * 1800 / pi + 3600  + TransformerPhaseShift - PhSh09 - 15) % 3600;

      CH08_Amplitude = round(sqrt(float(CH08_Sin) * float(CH08_Sin) + float(CH08_Cos) * float(CH08_Cos)) * CalCh08);
      CH08_Phase = round(atan2(CH08_Sin, CH08_Cos) * 1800 / pi + 3600  + TransformerPhaseShift - PhSh08 - 0) % 3600;

      CH08_Sin = CH09_Sin = CH10_Sin = CH11_Sin = CH12_Sin = CH13_Sin = CH14_Sin = 0;
      CH08_Cos = CH09_Cos = CH10_Cos = CH11_Cos = CH12_Cos = CH13_Cos = CH14_Cos = 0;

      // Now put the values in the payload string
      // The amplitudes will never be larger than some 300 (deciAmperes) and are always positive, so 3 digits wil do.
      // The angles will never be larger than 360 Degrees and are always positive, so 3 digits will do.
      sprintf(Payload, "%4d%4d%4d%4d%4d%4d%4d%4d%4d%4d%4d%4d%4d%4d",
        CH14_Amplitude, CH13_Amplitude, CH12_Amplitude, CH11_Amplitude, CH10_Amplitude, CH09_Amplitude, CH08_Amplitude,
        CH14_Phase / 10, CH13_Phase / 10, CH12_Phase / 10, CH11_Phase / 10, CH10_Phase / 10, CH09_Phase / 10, CH08_Phase / 10);
      DoPublishPower = true;
      
      if (false) //(ServiceMode)
      {
        Serial.print ("Currents and Phase: ");
        Serial.println (Payload);
      }
    }
  }
#endif
}

//********************************************************************************************************
void InitCurrents (void)
{
  Serial.println("Initializing Current Measurements....");
#ifdef USE_CURRENTS
  // Setup AD converter for 10 bit conversions for channels A15 .. A8
  ADCSRA = 0x84; // Enable the converter, clock div 16 for 1 MHz input clock
  // At a conversion rate of 20 kHz we have 50 ticks available for the conversion
  // The converter normally needs 13 clock ticks for 1 conversion.
  ADCSRB = 0x08; // Mux bit 5 = 1 for channels 8 .. 15;
  ADMUX = 0x47;  // Mux channel 15, Ref = AVCC
  DIDR2 = 0xFF;  // disable digital input buffers for channels 8 .. 15
  // start a conversion with ADCSRA |= 0x40;
  // read the result with: int ADresult = ADC;

  // We start at an unknown moment in the mains period and wait for a positive zero crossing.
  SampleNumber = 311;
  Serial.println("Current Measurements Initialized.");
  Serial.println();
#else
  Serial.println("No Current Measurements Configured.");
  Serial.println("Falling back to default Task Dispatcher.");
  Serial.println();
  HaveMainsPower = false;
#endif

  // Configure 16 bit Timer 3 for 20 kHz interrupt frequency = 50 usec interval
  TCCR3A = 0x00; // WaveformMode 4 = CTC
  TCCR3B = 0x09; // Prescaler 1, WaveformMode 4 = CTC, internal clock 16 MHz.
  TIMSK3 = 0x02; // Interrupt on OCRA match
  OCR3A = 799;   // 16 MHz div 800 = 20 kHz
}
