/////////////////////////////////////////////////////////////////////////////////////////
// Module Name: Sensors.cpp
// Project: PowerMonitor #42
// Routines to request data from several sensors
// by Jan Breemer
// jan@breem.nl, janbee@hack42.nl
//
// Functions in this module:
//
// UpdatePublicationScheduler ()
// UpdateSensors ()
// SearchDallasSensors ()
// ReadDallasSensor ()
// InitSensors ()
/////////////////////////////////////////////////////////////////////////////////////////

#include "Globals.h"

// For BME280 Temperature / Hygro / Baro sensor:
#include <Adafruit_BME280.h>
Adafruit_BME280 bme;
bool BME280_IsInitialized;
int Temperature, Hygro, Baro;

// For CCS811 CO2 / VOC sensor:
#include "Adafruit_CCS811.h"
Adafruit_CCS811 ccs;
bool CCS811_IsInitialized;
int eCO2, eVOC;

// For SDS011 Particle Sensor:
bool SDS011_IsInitialized;
int P2p5, P10, Measured_P2p5, Measured_P10;
byte ParticleBuffer[10];

// For Dallas DS 18x20 Temperature sensors:
#include <OneWire.h>
OneWire dsT(3); // pin 3, PORT E4
byte DallasCodes[MAX_DALLAS_SENSORS][8];
int DallasTemperatures[MAX_DALLAS_SENSORS];
byte NumberOfDallasSensors;
bool DoPublishDallasTemperatures;

// For sensors general:
bool InitialisationError;
bool DoPublishAirQuality;

// ********************************************************************************************************
void UpdatePublicationScheduler(void) // called at 1 Hz from UpdateSensors()
{
  static byte PubTime;
  static unsigned int Seconds;
  DIAGPIN_6_T

  switch (PubTime++)
  {
    case 0:
      DoPublishStatus = true;
      DoPublishAirQuality = true;
      DoPublishDallasTemperatures = true;
    break;

    case 10:
      if (ServiceMode) PubTime = 0; // 10 seconds
    break;

    case 59:
      PubTime = 0; // 1 minute
    break;      
  }
  
  switch (Seconds++)
  {
    case 0:       
      SDS_POWER_ON
    break;
    
    // give it some time to start-up and do measurements

    case 10:
      // Topic AirQuality will show the latest measured values.
      P10 = Measured_P10;
      P2p5 = Measured_P2p5;
      SDS_POWER_OFF
    break;

    case 600: // once per 10 minutes
      Seconds = 0;
    break;
  }

  static bool PrevButton1Pressed, PrevButton2Pressed;

  if (RequestButton1Pressed != PrevButton1Pressed) 
  {
    PrevButton1Pressed = RequestButton1Pressed;
    DoPublishStatus = true;
  }
  if (RequestButton2Pressed != PrevButton2Pressed) 
  {
    PrevButton2Pressed = RequestButton2Pressed;
    DoPublishStatus = true;
  }
}

// ********************************************************************************************************
void UpdateSensors(void) // Called from /FW_Monitor.ino/loop() at ca. 1 kHz
{
  static byte SNumber, DallasSensorNumber;
  byte DallasData[12], CCS811result;
  float Celsius;
  int raw;

  if (InitialisationError) if (Blink) HEARTBEATLED_ON else HEARTBEATLED_OFF

  // We need to read the data from the particle sensor every time it's available,
  // otherwise the Serial3 input buffer will be filled up,
  // which causes the data to be corrupted.
  // The particle sensor talks once every second when it is switched on.
  if (Serial3.available())
  {
    Serial3.readBytes(ParticleBuffer, 10);
    Measured_P2p5 = (ParticleBuffer[3] << 8) | ParticleBuffer[2];
    Measured_P10 =  (ParticleBuffer[5] << 8) | ParticleBuffer[4];
    if (false)
    {
      Serial.print (" Header: ");
      Serial.print (ParticleBuffer[0]);
      Serial.print (" Cmndr: ");
      Serial.print (ParticleBuffer[1]);
      Serial.print (" P10: ");
      Serial.print (Measured_P10);
      Serial.print (" P2p5: ");
      Serial.println (Measured_P2p5);
    }
  }

  if (DoReadaSensor) // set at a 5Hz, 200 ms rate by /Currents.cpp/ISR(TIMER3_COMPA_vect)
  {
    DIAGPIN_5_H
    DoReadaSensor = false;

    static byte SecondsCounter;
    if (++ SecondsCounter > 4)
    {
      SecondsCounter = 0;
      UpdatePublicationScheduler(); // once per second
    }

    switch (SNumber++)
    {
      case 0:
        EvaluateJumpers(false); // Do not show the results
      break;

      case 1:
      #ifdef USE_BME280
        Temperature = round(bme.readTemperature() * 10); // deci degrees C
        if (false) //(ServiceMode)
        {
          Serial.print ("BME280 Temperature = ");
          Serial.println (Temperature / 10);
        }
      #endif
      break;

      case 2:
      #ifdef USE_BME280
        Hygro = round(bme.readHumidity()); // % relative
        if (false) //(ServiceMode)
        {
          Serial.print ("BME280 Hygro = ");
          Serial.println (Hygro);
        }
      #endif
      break;

      case 3:
      #ifdef USE_BME280
        Baro = round(bme.readPressure() / 100.0); // mBar = HectoPascal
        if (false) //(ServiceMode)
        {
          Serial.print ("BME280 Barometric Pressure = ");
          Serial.println (Baro);
        }
      #endif
      break;

      case 4:
      #ifdef USE_CCS811
        CCS811result = ccs.readData(); // do the measurements
        if (false) //(ServiceMode)
        {
          Serial.print ("CCS811 result = ");
          Serial.println (CCS811result);
        }
      #endif
      break;

      case 5:
      #ifdef USE_CCS811
        eCO2 = -999; // for when it does not work
        if (CCS811_IsInitialized) eCO2 = ccs.geteCO2(); // in ppm
        if  (false) //(ServiceMode)
        {
          Serial.print ("CO2 = ");
          Serial.println (eCO2);
        }
      #endif
      break;

      case 6:
      #ifdef USE_CCS811
        eVOC = -999; // for when it does not work
        if (CCS811_IsInitialized) eVOC = ccs.getTVOC(); // in ppm
        if  (false) //(ServiceMode)
        {
          Serial.print ("VOC = ");
          Serial.println (eVOC);
        }
      #endif
      break;

      case 7: // Read a Dalsa Sensor
      #ifdef USE_DALLAS_SENSORS
        DIAGPIN_7_H
        if (false) //(ServiceMode)
        {
          Serial.print("Reading Dallas Sensor: ");
          Serial.println(DallasSensorNumber);
        }
        dsT.reset();
        dsT.select(DallasCodes[DallasSensorNumber]);
        dsT.write(0xBE); // read Scratchpad
        for (byte i = 0; i < 9; i++) DallasData[i] = dsT.read(); // we need 9 bytes
        raw = (DallasData[1] << 8) | DallasData[0];
        // DS18B20 has family code 40 (first number, decimal) and calibration is
        // T (Celcius) = readout * 0.0625.
        // DS18S20 has family code 16 (first number, decimal) and calibration is
        // T (Celcius) = readout * 0.5
        if (DallasCodes[DallasSensorNumber][0] == 0x28)
          Celsius = raw * 0.0625;
        else if (DallasCodes[DallasSensorNumber][0] == 0x28)
          Celsius = raw * 0.5;
        if (DallasSensorNumber < NumberOfDallasSensors)
          DallasTemperatures[DallasSensorNumber] = round(Celsius * 10);
        else
          DallasTemperatures[DallasSensorNumber] = -999;
        DIAGPIN_7_L
      #endif
      break;

      case 8: // Start conversion for the next Dallas Sensor
      #ifdef USE_DALLAS_SENSORS
        DIAGPIN_7_H
        if (false) //(ServiceMode)
        {
          Serial.print("Start conversion for Dallas Sensor: ");
          Serial.println(DallasSensorNumber);
        }        
        if (++DallasSensorNumber >= MAX_DALLAS_SENSORS) DallasSensorNumber = 0;
        dsT.reset();
        dsT.select(DallasCodes[DallasSensorNumber]);
        dsT.write(0x44, 1);  // start conversion command
        DIAGPIN_7_L
      #endif
      break;

      case 10:
        SNumber = 0; // start over again
      break;
    }
    DIAGPIN_5_L
  }
}

//*****************************************************************************
// Called at startup from InitSensors() or
// when a 'D' is entered in ServiceMode
void SearchDallasSensors (void) 

{
  byte i;
  Serial.println("Searching for Dallas Temperature Sensors....");
  NumberOfDallasSensors = 0;
  while (NumberOfDallasSensors < MAX_DALLAS_SENSORS)
  {
    if (!dsT.search(DallasCodes[NumberOfDallasSensors]))
    {
      Serial.print("Found ");
      Serial.print(NumberOfDallasSensors);
      Serial.println(" Dallas Temperature Sensors.");
      dsT.reset_search();
      return (NumberOfDallasSensors);
    }
    else
    {
      Serial.print("Sensor no ");
      Serial.print(NumberOfDallasSensors);
      Serial.print(": ROM = ");
      for (i = 0; i < 8; i++)
      {
        Serial.print(DallasCodes[NumberOfDallasSensors][i]);
        Serial.write(' ');
      }
      if (OneWire::crc8 (DallasCodes[NumberOfDallasSensors], 7) != DallasCodes[NumberOfDallasSensors][7])
      {
        Serial.println ("CRC is not valid!");
        InitialisationError = true;
        return;
      }
      switch (DallasCodes[NumberOfDallasSensors][0])
      {
        case 0x10:
          Serial.println (", Chip = DS18S20"); // or old DS1820. 0.5 K resolution
          break;
        case 0x28:
          Serial.println (", Chip = DS18B20"); // 0.0625 K resolution
          break;
        case 0x22:
          Serial.println (", Chip = DS1822");  // 0.0625 K resolution, +/- 2K accuracy
          break;
        default:
          Serial.println ("Device is not a DS18x20 family device.");
          InitialisationError = true;
          return;
      }
      NumberOfDallasSensors++;
    }
  }
}

//*****************************************************************************
// This function is only called at startup from InitSensors() or 
// when a 'D' is entered in ServiceMode
// During normal operation we use a method where the actions are
// divided over different timeslices.
void ReadDallasSensor (byte SensorNumber) // Called at startup from InitSensors()
{
  byte i, data[12];
  float celsius;

  dsT.reset();
  dsT.select(DallasCodes[SensorNumber]);
  dsT.write(0x44, 0);  // start conversion, with parasite power OFF at the end

  delay(1000);     // maybe 750ms is enough, maybe not

  dsT.reset();
  dsT.select(DallasCodes[SensorNumber]);
  dsT.write(0xBE);  // Read Scratchpad
  for (i = 0; i < 9; i++) data[i] = dsT.read(); // we need 9 bytes
  int16_t raw = (data[1] << 8) | data[0];
  // DS18B20 has family code 0x28 (first number, hex) and calibration is
  // T (Celcius) = readout * 0.0625.
  // DS18S20 has family code 0x10 (first number, hex) and calibration is
  // T (Celcius) = readout * 0.5
  if (DallasCodes[SensorNumber][0] == 0x28)
    celsius = (float)raw * 0.0625;
  else if (DallasCodes[SensorNumber][0] == 0x28)
    celsius = (float)raw * 0.5;
  Serial.print("Sensor no ");
  Serial.print(SensorNumber);
  Serial.print(": Temperature = ");
  Serial.print(celsius);
  Serial.println(" Celsius, ");
}

// ********************************************************************************************************
void InitSensors (void) // Called at startup by /FW_Monitor.ino/setup()
{
  delay (100); // give Current Measurement system time to report
  Wire.begin(); // initialize I2C system
  Serial.println("Start initialisation of sensors:");

#ifdef USE_BME280
  Serial.print("Initialize BME280 sensor for Temperature, Hygro and Baro.....");
  BME280_IsInitialized = true;
  if (!bme.begin())
  {
    Serial.println("Could not find a valid BME280 sensor, check wiring!");
    BME280_IsInitialized = false;
    InitialisationError = true;
  }
  Serial.println("OK");
#else
  Serial.println("BME280 sensor for Temperature, Hygro and Baro is not used.");
#endif

#ifdef USE_CCS811
  Serial.print("Initialize CCS811 sensor for eCO2 and eVOC..... ");
  CCS811_IsInitialized = true;
  if (!ccs.begin())
  {
    Serial.println("Could not find a valid CCS811 sensor, check wiring!");
    CCS811_IsInitialized = false;
    InitialisationError = true;
  }
  // Wait for the sensor to be ready
  unsigned int CCS811_TimeOut = 0;
  while (!ccs.available())
  {
    delay(1);
    if (++CCS811_TimeOut > 5000)
    {
      Serial.println("CCS811 sensor did not initialize in time");
      CCS811_IsInitialized = false;
      InitialisationError = true;
      break;
    }
  }
  Serial.println("OK");
#else
  Serial.println("CSS811 sensor for eCO2 and eVOC is not used.");
#endif

#ifdef USE_SDS011
  Serial.print("Initialize SDS011 Particle Sensor.....");
  SDS011_IsInitialized = true;
  Serial3.begin(9600);
  DDRE = 0x10;   // Power for the Particle Sensor initially OFF
  SDS_POWER_ON
  // the only way to find out if it's working is once we receive data
  unsigned int SDS011_TimeOut = 0;
  while (!Serial3.available()) // SDS should talk once every second
  {
    delay(1);
    if (++SDS011_TimeOut > 2000) // 2 seconds
    {
      Serial.print("No response from SDS011 Particle detector: ");
      Serial.println(SDS011_TimeOut);
      SDS011_IsInitialized = false;
      InitialisationError = true;
      break;
    }
  }
  delay(1); // ensure that the whole message is in the read buffer
  Serial3.readBytes(ParticleBuffer, 10); // empty this buffer
  Serial.println("OK");
  SDS_POWER_OFF
#else
  PORTE |= 0x10;  // high = disable power to the SDS011 unit
  Serial.println("SDS011 Particle sensor is not used.");
#endif

#ifdef USE_DALLAS_SENSORS
  byte i;
  for (i = 0; i < MAX_DALLAS_SENSORS; i++)
    DallasTemperatures[i] = -999;
  SearchDallasSensors();
  for (i = 0; i < NumberOfDallasSensors; i++)
    ReadDallasSensor(i);
  Serial.println();
  Serial.println("Enter 'D' to repeat the Dallas Sensor Search");
  Serial.println();
#else
  Serial.println("Dallas Temperature Sensors are not used.");
#endif

  if (InitialisationError)
    Serial.print("There were Errors");
  else
    Serial.print("No errors found");
  Serial.println(" during initialisation");
  Serial.println();
}
